#include <stdio.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <sys/un.h>
#include <signal.h>
#include <errno.h>
#include <sys/eventfd.h>

#include "mctd.h"
#include "mctd-log.h"
#include "mctd-mux.h"
#include "mctd-serial.h"
#include "mctd-modem-control.h"
#include "mctd-thread.h"
#include "mctd-thread-queue.h"

volatile sig_atomic_t _terminated = 0;

static void terminate_sig_handler(__unused int signum)
{
    _terminated = 1;
    mctd_remove_pid_file();
    mctd_serial_remove_own_lock_file();
}

static bool set_sig_handler(int sig_list[], ssize_t num, __sighandler_t handler)
{
    struct sigaction sa;
    int i;
    bool result = true;

    memset(&sa, 0, sizeof(sa));
    sa.sa_handler = handler;
    sa.sa_flags = SA_SIGINFO;

    for(i = 0; i < num; i++)
        if (sigaction(sig_list[i], &sa, NULL) < 0) {
            perror("sigaction");
            result = false;
        }

    return result;
}

static int receive_request(int fd, unsigned int *cmd_id, char **param)
{
    char buf[256];
    unsigned int param_len;
    MCTD_REQUEST_T request;

    /* recive header */
    if (mctd_receive_data(fd, buf, strlen(MCTD_HEADER), 3) != MCTD_RESULT_OK)
        return MCTD_RESULT_ERR_SOCKET;

    /* check header */
    if (!mctd_is_valid_header(buf))
        return MCTD_RESULT_ERR_SOCKET;

    /* recive fix parameter */
    if (mctd_receive_data(fd, buf, sizeof(MCTD_REQUEST_T), 3) != MCTD_RESULT_OK)
        return MCTD_RESULT_ERR_SOCKET;

    memcpy(&request, buf, sizeof(MCTD_REQUEST_T));

    *cmd_id = ntohl(mctd_u8tou32(request.cmd_id));
    param_len = ntohl(mctd_u8tou32(request.param_len));

    /* receive parameter if exsist */
    if (param_len == 0)
        return MCTD_RESULT_OK;

    *param = malloc(param_len+1);
    if (!*param) {
        perror("malloc");
        return MCTD_RESULT_ERR_MEMORY;
    }

    memset(*param, 0, param_len+1);
    if (mctd_receive_data(fd, *param, param_len, 3) != MCTD_RESULT_OK)
        return MCTD_RESULT_ERR_SOCKET;

    return MCTD_RESULT_OK;
}

static int send_response(int fd, const unsigned int cmd_id,
                                const unsigned int result, const char *param)
{
    MCTD_RESPONSE_T response;

    /* write header */
    if (write(fd, MCTD_HEADER, strlen(MCTD_HEADER)) == -1) {
        perror("write header.");
        return MCTD_RESULT_ERR_SOCKET;
    }

    /* write fix parameter */
    mctd_u32tou8(htonl(cmd_id), response.cmd_id);
    mctd_u32tou8(htonl(result), response.result);
    if (param)
        mctd_u32tou8(htonl(strlen(param)), response.param_len);
    else
        mctd_u32tou8(0, response.param_len);

    if (write(fd, &response, sizeof(response)) == -1) {
        perror("write fix parameter.");
        return MCTD_RESULT_ERR_SOCKET;
    }

    /* write parameter if exist */
    if (!param)
        return MCTD_RESULT_OK;

    if (write(fd, param, strlen(param)) == -1) {
        perror("write parameter.");
        return MCTD_RESULT_ERR_SOCKET;
    }

    return MCTD_RESULT_OK;
}

static void shutdown_peer(int fd)
{
    shutdown(fd, SHUT_RDWR);

    /* wait EOF, 5 seconds */
    mctd_wait_data(fd, 5);

    close(fd);
}

static int create_domain_socket()
{
    int sock;

    sock = socket(AF_UNIX, SOCK_STREAM, 0);
    if (sock == -1) {
        log_err("open socket failed for domain socket");
        return -1;
    }

    struct sockaddr_un sa = {0};
    sa.sun_family = AF_UNIX;
    strcpy(sa.sun_path, SOCK_PATH);

    remove(sa.sun_path);

    if (bind(sock, (struct sockaddr*) &sa, sizeof(struct sockaddr_un)) == -1) {
        log_err("bind failed for domain socket");
        close(sock);
        return -1;
    }

    if (listen(sock, 8) == -1) {
        log_err("listen failed for domain socket");
        close(sock);
        return -1;
    }

    return sock;
}

static bool has_arg(int argc, char *argv[], const char *param)
{
    int i;
    for (i = 1; i < argc; i++) {
        if ((strcmp(argv[i], param) == 0))
            return true;
    }
    return false;
}

static bool has_arg_poweron(int argc, char *argv[])
{
    return has_arg(argc, argv, "poweron");
}

static bool has_arg_no_at_command_log(int argc, char *argv[])
{
    return has_arg(argc, argv, "no-at-command-log");
}

static int max(int a, int b)
{
    return a > b ? a : b;
}

static void response_management(int fd, unsigned int cmd_id, int result, char *param)
{
    if (fd > -1) {
        send_response(fd, cmd_id, result, param);
        shutdown_peer(fd);
    }
}

static void queue_clean(int req_result)
{
    int fd;
    unsigned int cmd_id;
    char *req_param = NULL;
    char *res_param = NULL;
    int result;

    /* response clean */
    while (1) {
        if (mctd_thread_queue_res_dequeue(&fd, &cmd_id, &res_param, &result)
                                                                == MCTD_RESULT_OK)
            response_management(fd, cmd_id, result, res_param);
        else
            break;
    }

    /* request clean */
    while (1) {
        if (mctd_thread_queue_req_dequeue(&fd, &cmd_id, &req_param) == MCTD_RESULT_OK)
            response_management(fd, cmd_id, req_result, NULL);
        else
            break;
    }
}

static int poweroff_management(int fd, unsigned int cmd_id)
{
    if (mctd_thread_is_processing()) {

        /* modem control thread exit */
        mctd_thread_force_quit();

        /* queue clean */
        queue_clean(MCTD_RESULT_ERR_INTERRUPT);

        /* serial refresh */
        mctd_serial_force_close();

        /* modem control thread restart */
        if (mctd_thread_create()) {
            log_err("thread create error.");
            return -1;
        }
    }

    /* command enqueue */
    if (mctd_thread_queue_req_enqueue(fd, cmd_id, NULL) == MCTD_RESULT_ERR_MEMORY) {
        log_err("malloc failed.");
        return -1;
    }

    return 0;
}

static int init_command_request(int efd, bool *error_end, int mctd_cmd)
{
    int fd;
    int result;
    unsigned int cmd_id;
    char *res_param = NULL;
    struct timeval tv;
    fd_set readfds;
    uint64_t r64;
    ssize_t rlen;

    FD_ZERO(&readfds);
    FD_SET(efd, &readfds);
    tv.tv_sec = 60;
    tv.tv_usec = 0;

    if (mctd_thread_queue_req_enqueue(-1, mctd_cmd, NULL)
                                                    == MCTD_RESULT_ERR_MEMORY) {
        log_err("malloc failed.");
        return -1;
    }
    result = select(efd + 1, &readfds, NULL, NULL, &tv);

    if (result == 0) {
        log_err("poweron failed.");
        return -1;
    }
    if (result == -1) {
        perror("select");
        *error_end = true;
        return -1;
    }

    do {
        rlen = read(efd, &r64, sizeof(r64));
    } while (errno == EINTR && rlen < 0);
    if (mctd_thread_queue_res_dequeue(&fd, &cmd_id, &res_param, &result)
                                                    == MCTD_RESULT_ERR_MEMORY) {
        log_err("malloc failed.");
        return -1;
    }
    free(res_param);
    res_param = NULL;

    return 0;
}

/**
 * @param argv[1]: "poweron" with powered-on Armadillo.
 */
int main(int argc, char *argv[])
{
    int sock = -1;
    int fd;
    int result;
    int val;
    bool error_end = false;
    unsigned int cmd_id;
    char *req_param = NULL;
    char *res_param = NULL;
    struct timeval tv;
    fd_set readfds;
    int register_interval_sec;
    uint64_t r64;
    ssize_t rlen;
    int efd;

    int terminate_sig_list[] = {
            SIGHUP, SIGINT, SIGSEGV, SIGQUIT, SIGTERM
        };

    syslog_open("aiot-modem-controld");
    log_info("start aiot-modem-controld.");

    if (mctd_has_pid_file()) {
        log_warn("has pid file.");
        syslog_close();
        return 0;
    }

    if (!set_sig_handler(terminate_sig_list,
                sizeof(terminate_sig_list) / sizeof(terminate_sig_list[0]),
                terminate_sig_handler)) {
        log_err("set_sig_handler failed");
        error_end = true;
        goto bail;
    }

    mkdir(DAEMON_PATH, 0777);

    sock = create_domain_socket();
    if (sock == -1) {
        log_err("create domain socket failed.");
        error_end = true;
        goto bail;
    }

    if (!mctd_create_pid_file()) {
        log_warn("mctd_create_pid_file failed.");
        error_end = true;
        goto bail;
    }

    /* thread eventfd */
    efd = eventfd(0, EFD_CLOEXEC | EFD_NONBLOCK);
    if (efd < 0) {
        log_err("create eventfd failed.");
        error_end = true;
        goto bail;
    }
    mctd_thread_set_efd(efd);

    set_output_at_command_log(!has_arg_no_at_command_log(argc, argv));

    if (mctd_thread_create()) {
        log_err("thread create error.");
        error_end = true;
        goto bail;
    }

    if (has_arg_poweron(argc, argv)) {
        /* power off ems31-j, for unexpected reboot. */
        if(init_command_request(efd, &error_end, MCTD_CMD_POWER_SUPPLY_OFF))
            goto bail;

        /* start with powered-on Armadillo */
        if(init_command_request(efd, &error_end, MCTD_CMD_POWERON))
            goto bail;
    }

    if (mctd_thread_queue_req_enqueue(-1, MCTD_CMD_FIRST_REGISTATION, NULL)
                                                        == MCTD_RESULT_ERR_MEMORY) {
        log_err("malloc failed.");
        goto bail;
    }

    while (!_terminated) {
        FD_ZERO(&readfds);
        FD_SET(sock, &readfds);
        FD_SET(efd, &readfds);

        register_interval_sec = mctd_modem_control_get_check_register_interval_sec();
        if (register_interval_sec > 0) {
            tv.tv_sec = register_interval_sec;
            tv.tv_usec = 0;

            result = select(max(sock, efd) + 1, &readfds, NULL, NULL, &tv);

            if (result == 0) {
                if (!mctd_modem_control_is_suspend()) {
                    /* check register */
                    if (!mctd_thread_is_processing())
                        if (mctd_thread_queue_req_enqueue(-1, MCTD_CMD_REGISTATION, NULL)
                                                                == MCTD_RESULT_ERR_MEMORY) {
                            log_err("malloc failed.");
                            goto bail;
                        }
                }
                continue;
            }
        } else {
            result = select(max(sock, efd) + 1, &readfds, NULL, NULL, NULL);

            if (result == 0)
                continue;
        }

        if (result == -1) {
            perror("select");
            error_end = true;
            goto bail;
        }

        if (FD_ISSET(sock, &readfds)) {
            fd = accept(sock, NULL, NULL);
            if (fd == -1) {
                perror("accept");
                log_warn("accept failed for domain socket");
                error_end = true;
                goto bail;
            }

            /* set nonblock */
            val = 1;
            ioctl(fd, FIONBIO, &val);

            if (receive_request(fd, &cmd_id, &req_param) != MCTD_RESULT_OK) {
                log_warn("failed to receive_request");
                shutdown_peer(fd);
                continue;
            }

            if (!mctd_modem_control_has_cmd_id(cmd_id)) {
                response_management(fd, cmd_id, MCTD_RESULT_ERR_CMD_NOT_FOUND, NULL);
                shutdown_peer(fd);
            } else {
                switch(cmd_id) {
                case MCTD_CMD_POWEROFF:
                case MCTD_CMD_POWEROFF_AND_FIN:
                    if (poweroff_management(fd, cmd_id)) {
                        /* Poweroff Management failed. */
                        response_management(fd, cmd_id, MCTD_RESULT_ERR_PROCESS, NULL);
                        error_end = true;
                        goto bail;
                    }

                    break;

                case MCTD_CMD_FIN:
                    _terminated = 1;
                    response_management(fd, cmd_id, MCTD_RESULT_OK, NULL);
                    break;

                default:
                    result = mctd_thread_queue_req_enqueue(fd, cmd_id, req_param);
                    if (result == MCTD_RESULT_ERR_CMD_BUSY)
                        /* Command queue Full*/
                        response_management(fd, cmd_id, MCTD_RESULT_ERR_CMD_BUSY, NULL);
                    else if (result == MCTD_RESULT_ERR_MEMORY) {
                        log_err("malloc failed.");
                        goto bail;
                    }

                    break;
                }
            }

            free(req_param);
            req_param = NULL;

        }

        if (FD_ISSET(efd, &readfds)) {
            do {
                rlen = read(efd, &r64, sizeof(r64));
            } while (errno == EINTR && rlen < 0);

            if (mctd_thread_queue_res_dequeue(&fd, &cmd_id, &res_param, &result)
                                                        == MCTD_RESULT_ERR_MEMORY) {
                log_err("malloc failed.");
                goto bail;
            }
            response_management(fd, cmd_id, result, res_param);
            if (cmd_id == MCTD_CMD_POWEROFF_AND_FIN) {
                _terminated = 1;
            }
            free(res_param);
            res_param = NULL;

        }
    }

bail:
    log_info("finish aiot-modem-controld.");
    mctd_modem_control_finalize();
    mctd_thread_quit();
    queue_clean(MCTD_RESULT_ERR_DAEMON_QUIT);

    if (sock != -1)
        close(sock);

    if (efd != -1)
        close(efd);

    mctd_serial_remove_own_lock_file();
    mctd_remove_pid_file();
    syslog_close();

    return error_end?1:0;
}
